"""Module for ivector extractor training"""

from montreal_forced_aligner.ivector.trainer import (
    DubmTrainer,
    IvectorTrainer,
    TrainableIvectorExtractor,
)

__all__ = ["trainer", "DubmTrainer", "IvectorTrainer", "TrainableIvectorExtractor"]
